"""
Lightnet Network Module |br|
This module contains classes and functions to create deep neural network with pytorch_.
It is mostly targeted at Object detection networks, but contains the necessary tools to build any CNN type network.
"""

from . import layer
from . import loss
from . import module
